from os import environ
from collections import OrderedDict

from django import forms, VERSION
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase
from django.db import models
from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from django.template.base import Template, TemplateSyntaxError
from django.template.context import Context

from .models import InheritedModel
from .templatetags.model_meta import model_meta_verbose_name, model_meta_verbose_name_plural
from .templatetags.gravatar import gravatar_get_url, gravatar_get_img
from .toolbox import set_form_widgets_attrs, choices_list, get_choices, get_site_url, get_model_class_from_string, \
    get_model_class_from_settings, import_app_module, import_project_modules


class MyForm(forms.Form):

    field1 = forms.CharField(label='f1', max_length=100)
    field2 = forms.CharField(label='f2', max_length=100)


class MyParentModel(models.Model):

    code = models.CharField('dummy', max_length=64)
    expired = models.BooleanField('Expired', help_text='dummy', default=False)

    class Meta:
        abstract = True


class MyChildModel1(InheritedModel, MyParentModel):

    class Fields:
        code = 'Secret code'
        expired = {'help_text': 'This code is expired.'}

    class Meta:
        verbose_name = 'Verb'
        verbose_name_plural = 'VerbPlural'


class MyChildModel2(InheritedModel, MyParentModel):

    class Fields:
        code = 'Non-secret code'


######################################################################################

class EtcTestCase(TestCase):

    @classmethod
    def render(cls, string, context):
        return Template(string).render(Context(context))


class InheritedModelTest(EtcTestCase):

    def test_texts(self):
        model1 = MyChildModel1()
        fields = {f.name: f for f in model1._meta.fields}
        self.assertEqual(fields['code'].verbose_name, 'Secret code')
        self.assertEqual(fields['expired'].help_text, 'This code is expired.')

        model1 = MyChildModel2()
        fields = {f.name: f for f in model1._meta.fields}
        self.assertEqual(fields['code'].verbose_name, 'Non-secret code')
        self.assertEqual(fields['expired'].help_text, 'dummy')


class ModelMetaTemplateTagsTest(EtcTestCase):

    def test_verbose_name_singular(self):
        m = MyChildModel1
        self.assertEqual(model_meta_verbose_name(m), 'Verb')

    def test_verbose_name_plural(self):
        m = MyChildModel1
        self.assertEqual(model_meta_verbose_name_plural(m), 'VerbPlural')


class ModelFieldTemplateTagsTest(EtcTestCase):

    def test_model_field_verbose_name(self):
        result = self.render(
            "{% load model_field %}{% model_field_verbose_name from model.first_name %}",
            {'model': User()}
        )
        self.assertEqual(result, 'first name')

        #
        # `from` missing
        self.assertRaises(
            TemplateSyntaxError,
            self.render,
            "{% load model_field %}{% model_field_verbose_name model.first_name %}", {}
        )

        #
        # `as` clause
        context = {'model': User()}
        result = self.render(
            "{% load model_field %}{% model_field_verbose_name from model.first_name as a %}",
            context
        )
        self.assertEqual(result, '')
        self.assertEqual(context['a'], 'first name')

        #
        # Wrong model-field delimiter.
        result = self.render(
            "{% load model_field %}{% model_field_verbose_name from model-first_name %}",
            context
        )
        self.assertEqual(result, '')

        with self.settings(DEBUG=True):
            self.assertRaises(
                TemplateSyntaxError,
                self.render,
                "{% load model_field %}{% model_field_verbose_name from model-first_name %}", {}
            )

        #
        # No model in context.
        result = self.render(
            "{% load model_field %}{% model_field_verbose_name from model.first_name %}",
            {}
        )
        self.assertEqual(result, '')

        with self.settings(DEBUG=True):
            self.assertRaises(
                TemplateSyntaxError,
                self.render,
                "{% load model_field %}{% model_field_verbose_name from model.first_name %}", {}
            )

        #
        # No field.
        result = self.render(
            "{% load model_field %}{% model_field_verbose_name from model.unknown %}",
            context
        )
        self.assertEqual(result, '')

        with self.settings(DEBUG=True):
            self.assertRaises(
                TemplateSyntaxError,
                self.render,
                "{% load model_field %}{% model_field_verbose_name from model.unknown %}", context
            )

    def test_model_field_help_text(self):
        result = self.render(
            "{% load model_field %}{% model_field_help_text from model.is_staff %}",
            {'model': User()}
        )
        self.assertIn('whether the user can log', result)

        #
        # `from` missing
        self.assertRaises(
            TemplateSyntaxError,
            self.render,
            "{% load model_field %}{% model_field_help_text %}", {}
        )

        #
        # `as` clause
        context = {'model': User()}
        result = self.render(
            "{% load model_field %}{% model_field_help_text from model.is_staff as a %}",
            context
        )
        self.assertEqual(result, '')
        self.assertIn('whether the user can log', context['a'])

        #
        # Wrong model-field delimiter.
        result = self.render(
            "{% load model_field %}{% model_field_help_text from model-is_staff %}",
            context
        )
        self.assertEqual(result, '')

        with self.settings(DEBUG=True):
            self.assertRaises(
                TemplateSyntaxError,
                self.render,
                "{% load model_field %}{% model_field_help_text from model-is_staff %}", {}
            )

        #
        # No model in context.
        result = self.render(
            "{% load model_field %}{% model_field_help_text from model.is_staff %}",
            {}
        )
        self.assertEqual(result, '')

        with self.settings(DEBUG=True):
            self.assertRaises(
                TemplateSyntaxError,
                self.render,
                "{% load model_field %}{% model_field_help_text from model.is_staff %}", {}
            )

        #
        # No field.
        result = self.render(
            "{% load model_field %}{% model_field_help_text from model.unknown %}",
            context
        )
        self.assertEqual(result, '')

        with self.settings(DEBUG=True):
            self.assertRaises(
                TemplateSyntaxError,
                self.render,
                "{% load model_field %}{% model_field_help_text from model.unknown %}", context
            )


class FormTest(EtcTestCase):

    def test_set_form_widgets_attrs(self):
        f = MyForm()
        d = OrderedDict()
        d['class'] = 'clickable'
        d['data-a'] = lambda f: f.__class__.__name__
        set_form_widgets_attrs(f, d)
        output = f.as_p()

        self.assertIn('class="clickable" data-a="CharField" id="id_field1"', output)
        self.assertIn('class="clickable" data-a="CharField" id="id_field2"', output)


class GravatarTemplateTagsTest(EtcTestCase):

    def test_verbose_get_url(self):
        u = User(username='idle')
        url = gravatar_get_url(u, 101, 'retro')
        self.assertIn('http://www.gravatar.com/avatar/ec2f993aec2c27fc750119ab17b16cdb/', url)
        self.assertIn('retro', url)
        self.assertIn('101', url)

        u = User(email='idle@sign.som')
        url = gravatar_get_url(u, 101, 'retro')
        self.assertIn('http://www.gravatar.com/avatar/37e24208b31f2a8f1e0f84d4c93fdfb0/', url)
        self.assertIn('retro', url)
        self.assertIn('101', url)

        url = gravatar_get_url('idle@sign.som', 101, 'retro')
        self.assertIn('http://www.gravatar.com/avatar/37e24208b31f2a8f1e0f84d4c93fdfb0/', url)
        self.assertIn('retro', url)
        self.assertIn('101', url)

        self.assertEqual(gravatar_get_url(None), '')

    def test_verbose_get_img(self):
        u = User(username='idle')
        url = gravatar_get_img(u, 101, 'retro')
        self.assertIn('http://www.gravatar.com/avatar/ec2f993aec2c27fc750119ab17b16cdb/', url)
        self.assertIn('retro', url)
        self.assertIn('retro', url)
        self.assertIn('<img src="', url)

        self.assertEqual(gravatar_get_img(None), '')


class ChoicesTest(EtcTestCase):

    def test_choices(self):

        types_dict = choices_list(
            (1, 'T1'),
            (2, 'T2'),
        )

        self.assertEqual(len(types_dict), 2)
        self.assertEqual(types_dict[1], 'T1')
        self.assertEqual(types_dict[2], 'T2')

        self.assertEqual(list(types_dict.keys()), [1, 2])

        ch = get_choices(types_dict)
        self.assertEqual(len(ch), 2)

        self.assertEqual(ch[0][0], 1)
        self.assertEqual(ch[0][1], 'T1')

        self.assertEqual(ch[1][0], 2)
        self.assertEqual(ch[1][1], 'T2')


class GetModelClassTest(EtcTestCase):

    def test_from_settings(self):
        attr_name = 'some'
        fake_settings = type('fake', (object,), {attr_name: 'auth.User'})

        self.assertIs(get_model_class_from_settings(fake_settings, attr_name), User)

    def test_from_string(self):
        cl = get_model_class_from_string('auth.User')
        self.assertIs(cl, User)

        self.assertRaises(ImproperlyConfigured, get_model_class_from_string, 'some')
        self.assertRaises(ImproperlyConfigured, get_model_class_from_string, 'etc.InheritedModel')


class GetSiteUrlTest(EtcTestCase):

    def test_basic(self):
        self.assertEqual(get_site_url(), 'http://example.com')

        with self.settings(SITE_PROTO='htt'):
            self.assertEqual(get_site_url(), 'htt://example.com')
            environ['SITE_PROTO'] = 'ttp'
            self.assertEqual(get_site_url(), 'ttp://example.com')

        environ['SITE_PROTO'] = 'https'
        self.assertEqual(get_site_url(), 'https://example.com')

        environ['SITE_SCHEME'] = 'ftp'
        self.assertEqual(get_site_url(), 'https://example.com')

        del environ['SITE_PROTO']
        self.assertEqual(get_site_url(), 'ftp://example.com')

        environ['SITE_URL'] = 'http://pythonz.net'
        self.assertEqual(get_site_url(), 'http://pythonz.net')

        environ['SITE_DOMAIN'] = 'mydomain.loc'
        self.assertEqual(get_site_url(), 'ftp://mydomain.loc')

        environ['SITE_URL'] = 'http://pythonz.net'
        self.assertEqual(get_site_url(), 'ftp://mydomain.loc')

        del environ['SITE_DOMAIN']
        del environ['SITE_URL']
        del environ['SITE_SCHEME']

        class FakeRequest(object):

            scheme = 'xyz'

            @classmethod
            def get_host(cls):
                return 'fake'

        self.assertEqual(get_site_url(request=FakeRequest), 'xyz://example.com')

        if VERSION < (1, 7):
            Site._meta.installed = False
            self.assertEqual(get_site_url(), 'http://undefined-domain.local')

    def test_tempalte_tag(self):
        url = 'http://pythonz.net'

        environ['SITE_URL'] = url
        result = self.render('{% load etc_misc %}{% site_url %}', {})
        self.assertEqual(result, url)


class ImportModulesTest(EtcTestCase):

    def test_import_app_module(self):
        m = import_app_module('etc', 'utils')
        self.assertTrue(hasattr(m, 'ModelBase'))

        self.assertRaises(ImportError, import_app_module, 'unknown', 'uknown')

        m = import_app_module('etc', 'uknown')
        self.assertIsNone(m)

        m = import_app_module('django.contrib.admin.apps.SimpleAdminConfig', '__init__')
        self.assertTrue(hasattr(m, 'site'))

    def test_import_project_modules(self):
        m = import_project_modules('toolbox')

        self.assertTrue(len(m) == 1)
        self.assertTrue(hasattr(m[0], 'get_site_url'))



from django.db.models.base import ModelBase


class InheritedModelMetaclass(ModelBase):

    def __new__(cls, name, bases, attrs):
        cl = super(InheritedModelMetaclass, cls).__new__(cls, name, bases, attrs)
        texts_marker = '_texts_applied'
        if not getattr(cl, texts_marker, False):
            try:
                names_map = {f.name: f for f in cl._meta.fields}
            except AttributeError:
                return cl
            fields_cl = getattr(cl, 'Fields', None)
            if fields_cl is not None:
                for attr_name, val in fields_cl.__dict__.items():
                    if not attr_name.startswith('_'):
                        if attr_name in names_map:
                            field = names_map[attr_name]
                            if not isinstance(val, dict):
                                val = {'verbose_name': val}
                            for field_attr, field_val in val.items():
                                setattr(field, field_attr, field_val)
            setattr(cl, texts_marker, True)
        return cl


class DomainGetter(object):

    __slots__ = ['domain']

    def __init__(self, domain):
        self.domain = domain

    def get_host(self):
        return self.domain



from os import environ
from functools import partial
from collections import OrderedDict

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import module_has_submodule

try:
    from django.contrib.sites.shortcuts import get_current_site

except ImportError:  # Django < 1.7
    from django.contrib.sites.models import get_current_site

try:
    from django.apps import apps
    apps_get_model = apps.get_model

except ImportError:  # Django < 1.7
    from django.db.models import get_model
    apps_get_model = None

try:
    from django.utils.module_loading import import_module

except ImportError:
    # Django <=1.9.0
    from django.utils.importlib import import_module

from .utils import DomainGetter


def choices_list(*choices):
    """Helps to define choices for models, that could be addressed
    later as dictionaries.

    To be used in conjunction with `get_choices()`.

    Example:

        class MyModel(models.Model):

            TYPE_ONE = 1
            TYPE_TWO = 2

            TYPES = choices_list(
                (TYPE_ONE, 'Type one title'),
                (TYPE_TWO, 'Type two title'),
            )

            type = models.PositiveIntegerField('My type', choices=get_choices(TYPES), default=TYPE_TWO)

            def get_display_type(self):
                return self.TYPES[self.type]

    :param set|list choices:
    :rtype: OrderedDict
    :return: Choices ordered dictionary
    """
    return OrderedDict(choices)


def get_choices(choices_list):
    """Returns model field choices from a given choices list.
    Choices list is defined with `choices_list()`.

    :param OrderedDict choices_list:
    :rtype: tuple
    :return: Choices tuple
    """
    return tuple((key, val) for key, val in choices_list.items())


def set_form_widgets_attrs(form, attrs):
    """Applies a given HTML attributes to each field widget of a given form.

    Example:

        set_form_widgets_attrs(my_form, {'class': 'clickable'})

    """
    for _, field in form.fields.items():
        attrs_ = dict(attrs)
        for name, val in attrs.items():
            if hasattr(val, '__call__'):
                attrs_[name] = val(field)
        field.widget.attrs = field.widget.build_attrs(attrs_)


def get_model_class_from_string(model_path):
    """Returns a certain model as defined in a string formatted `<app_name>.<model_name>`.

    Example:

        model = get_model_class_from_string('myapp.MyModel')

    """
    try:
        app_name, model_name = model_path.split('.')
    except ValueError:
        raise ImproperlyConfigured('`%s` must have the following format: `app_name.model_name`.' % model_path)

    if apps_get_model is None:
        model = get_model(app_name, model_name)
    else:
        try:
            model = apps_get_model(app_name, model_name)
        except (LookupError, ValueError):
            model = None

    if model is None:
        raise ImproperlyConfigured('`%s` refers to a model `%s` that has not been installed.' % (model_path, model_name))

    return model


def get_model_class_from_settings(settings_module, settings_entry_name):
    """Returns a certain model as defined in a given settings module.

    Example:

        myapp/settings.py

            from django.conf import settings

            MY_MODEL = getattr(settings, 'MYAPP_MY_MODEL', 'myapp.MyModel')


        myapp/some.py

            from myapp import settings

            model = get_model_class_from_settings(settings, 'MY_MODEL')

    """
    return get_model_class_from_string(getattr(settings_module, settings_entry_name))


def get_site_url(request=None):
    """Tries to get a site URL from environment and settings
    in the following order:

    1. (SITE_PROTO / SITE_SCHEME) + SITE_DOMAIN
    2. SITE_URL
    3. Django Sites contrib
    4. Request object

    :param HttpRequest request: Request object to deduce URL from.
    :rtype: str

    """
    env = partial(environ.get)
    settings_ = partial(getattr, settings)

    domain = None
    scheme = None
    url = None

    for src in (env, settings_):
        if url is None:
            url = src('SITE_URL', None)

        if domain is None:
            domain = src('SITE_DOMAIN', None)

        if scheme is None:
            scheme = src('SITE_PROTO', src('SITE_SCHEME', None))

    if domain is None and url is not None:
        scheme, domain = url.split('://')[:2]

    if domain is None:
        site = get_current_site(request or DomainGetter(domain))
        domain = site.domain

    if scheme is None and request:
        scheme = request.scheme

    if domain is None:
        domain = 'undefined-domain.local'

    if scheme is None:
        scheme = 'http'

    domain = domain.rstrip('/')

    return '%s://%s' % (scheme, domain)


def import_app_module(app_name, module_name):
    """Returns a module from a given app by its name.

    :param str app_name:
    :param str module_name:
    :rtype: module or None

    """
    name_split = app_name.split('.')
    if name_split[-1][0].isupper():  # Seems that we have app config class path here.
        app_name = '.'.join(name_split[:-2])

    module = import_module(app_name)

    try:
        sub_module = import_module('%s.%s' % (app_name, module_name))
        return sub_module

    except:

        # The same bubbling strategy as in autodiscover_modules().
        if module_has_submodule(module, module_name):  # Module is in a package.
            raise

        return None


def import_project_modules(module_name):
    """Imports modules from registered apps using given module name
    and returns them as a list.

    :param str module_name:
    :rtype: list

    """
    from django.conf import settings

    submodules = []
    for app in settings.INSTALLED_APPS:
        module = import_app_module(app, module_name)
        if module is not None:
            submodules.append(module)

    return submodules



from django import template

from ..toolbox import get_site_url

register = template.Library()


@register.simple_tag(takes_context=True)
def site_url(context):
    """Tries to get a site URL from environment and settings.

    See toolbox.get_site_url() for description.

    Example:

        {% load etc_misc %}
        {% site_url %}

    """
    return get_site_url(request=context.get('request', None))



from django.utils import six

from .utils import InheritedModelMetaclass


class InheritedModel(six.with_metaclass(InheritedModelMetaclass)):
    """Mix in this class into target model (inherit from it) and define `Fields` class inside it
    to be able to customize field attributes (e.g. texts) of a base-parent model.


    Example:

        from etc.models import InheritedModel


        class MyParentModel(models.Model):

            code = models.CharField('dummy', max_length=64)
            expired = models.BooleanField('Expired', help_text='dummy')

            class Meta:
                abstract = True


        class MyChildModel1(InheritedModel, MyParentModel):  # NOTE: InheritedModel must go first.

            time_created = models.DateTimeField('Date created', auto_now_add=True)

            class Fields:  # Defining a class with fields custom fields data.
                code = 'Secret code'  # This is treated as verbose_name.
                expired = {'help_text': 'This code is expired.'}


        class MyChildModel2(InheritedModel, MyParentModel):

            code = models.CharField('dummy', max_length=128, unique=True, editable=False)

            class Fields:
                code = 'Non-secret code'
                expired = {'help_text': 'Do not check it. Do not.'}

    """
